﻿/*	Application		IPC receiver-side using socket (UDP)		
	File			Program.cs (IPC_socket_Receiver_with_Reply)
	Purpose			Demonstration of local IPC, using a blocking socket
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket, and binds to local address (uses INADDR_ANY)
					2. Waits for a message to arrive on its port (8007)
					3. Displays the message 
					4. Reverses the message contents
					5. Sends the modified message back as a reply to the sender (to port 8008)
					6. Exits once a key is pressed
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace IPC_socket_Receiver_with_Reply
{
    class C_IPC_socket_Receiver
    {
        public const Int32 RECEIVE_BUFFER_SIZE = 100;
        public const Int32 RECEIVE_PORT = 8007;
        public const Int32 REPLY_PORT = 8008;  // The port to send the reply to (the port the original sender will be listening on)
        static Encoding enc8 = Encoding.UTF8;

        Socket m_IPC_ReceiveSOCKET;
        Socket m_IPC_ReplySOCKET;
        public byte[] m_szRecvBuf;
        public byte[] m_szReplyBuf;
        IPEndPoint m_localIPEndPoint;   // Local IP address and port
        IPEndPoint m_SenderEndPoint;    // This will hold the address and port of the sender after a message has been received

        static void Main(string[] args)
        {
            Console.WriteLine("Local Inter-Process Communication (receiver side)");
            Console.WriteLine("This program waits for a message to arrive on port 8007 (decimal).");
            Console.WriteLine("The message is displayed, and the message contents are reversed.");
            Console.WriteLine("The modified message is sent back to the sender as a reply, on port 8008.");
            Console.WriteLine("The process exits when a key is pressed.");
            Console.WriteLine("--------------------------------------------------------------------------");
            Console.WriteLine("Waiting for message ...   ");

            C_IPC_socket_Receiver IPC_socket_Receiver = new C_IPC_socket_Receiver();
        }

        C_IPC_socket_Receiver()
        {
            string sMessageString;
            bool bSuccess;
	
	        // Create a socket - this is the logical 'communication endpoint'
	        bSuccess = CreateReceiveSocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }

	        // Create the socket that will be used for sending the reply message
	        bSuccess = CreateReplySocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }

	        // Set the Local Address details (IP address and  port number)
	        SetUpLocalAddressStruct();

	        // Bind (The operating system makes a mapping between this process and the requested Port)
	        bSuccess = BindToLocalAddress();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because binding to the requested port was not allowed (port may be in use)" );
		        CloseSocketAndExit();
	        }

	        // Wait for / receive the message
            sMessageString = ReadMessageFromReceiveBufferOrWait();
            Console.WriteLine("   Message received:       {0}", sMessageString);
	        Console.WriteLine("   from sender at address: {0}", m_SenderEndPoint.Address);
	
	        // Reverse the received message contents in the buffer
            sMessageString = ReverseString(sMessageString);
	        Console.WriteLine("   Reversed reply message: {0}", sMessageString);

	        // Send the reply message to the original message sender
            SendMessage(sMessageString);
	        Console.WriteLine("   Reply message sent");
	
	        // Close the sockets
	        CloseSocketAndExit();
        }
	
        bool CreateReceiveSocket()
        {
	        try
            {
                m_IPC_ReceiveSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }

        bool CreateReplySocket()
        {
	        try
            {
                m_IPC_ReplySOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }

        void SetUpLocalAddressStruct()
        {   // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(RECEIVE_PORT));
        }

        bool BindToLocalAddress()
        {   
            try
            {   // Bind to the local IP Address and selected port
                m_IPC_ReceiveSOCKET.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
		        Console.WriteLine("bind() Failed!");
                return false;
            }
        }

        string ReadMessageFromReceiveBufferOrWait()
        {
            string sReceivedStr = "";
            int iReceiveByteCount;
            try
            {
                m_szRecvBuf = new byte[RECEIVE_BUFFER_SIZE];
                m_SenderEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
                EndPoint SenderAddress = (EndPoint)m_SenderEndPoint;
                iReceiveByteCount = m_IPC_ReceiveSOCKET.ReceiveFrom(m_szRecvBuf, ref SenderAddress);
                
                // Place the sender's address into the m_SenderEndPoint, from the SenderAddress EndPoint
                m_SenderEndPoint.Address = IPAddress.Parse(((IPEndPoint)SenderAddress).Address.ToString());
                sReceivedStr = enc8.GetString(m_szRecvBuf, 0, iReceiveByteCount); // Convert the received message into string format
            }
            catch (SocketException ex)
            {
                Console.WriteLine("Receive failed");
                CloseSocketAndExit();
            }
            return sReceivedStr;
        }

        string ReverseString(string sStr)
        {
            if(sStr == null)
            {
                return null;
            }

            char[] cArray = sStr.ToCharArray(); // Convert the string to an array of characters
            Array.Reverse(cArray);              // Reverse the array
            return new String(cArray);          // Create a new string from the array and return it
        }

        void SendMessage(string sSendString)
        {
            // Send reply back to the original sender at the address they sent from, but set the reply port
            // to be the one the sender will use to receive on (which is not necessarily the one they sent from)
            m_SenderEndPoint.Port = REPLY_PORT;

            // Convert sSendString into the reply byte array format (!C# chars are two bytes each!)
             m_szReplyBuf = Encoding.ASCII.GetBytes(sSendString);
 
            try
            {
                
                m_IPC_ReplySOCKET.SendTo(m_szReplyBuf, m_SenderEndPoint);
            }       
            catch (SocketException ex)
            {
		        Console.WriteLine("sendto() Failed!");
		        CloseSocketAndExit();
	        }
        }

        void CloseSocketAndExit()
        {
            m_IPC_ReceiveSOCKET.Close();
            m_IPC_ReplySOCKET.Close();
 	        Console.WriteLine("Socket closed");
	        Console.WriteLine("(press a key to continue)");
	        Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            System.Environment.Exit(1);
        }
    }
}