﻿/*	Application		IPC sender-side using socket (UDP)		
	File			Program.cs (IPC_socket_Sender_With_reply)
	Purpose			Demonstration of local IPC, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
					2. Waits for user to type a message
					3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
					4. Waits for a reply message to be returned (on port 8008)
					5. Displays the reply message
					6. Exits once the reply message has been displayed
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace IPC_socket_Sender_with_Reply
{
    class C_IPC_socket_Sender
    {
        public const Int32 RECEIVE_BUFFER_SIZE = 100;
        public const Int32 SEND_PORT = 8007;
        public const Int32 RECEIVE_PORT = 8008;
        static Encoding enc8 = Encoding.UTF8;

        Socket m_IPC_SendSOCKET;
        Socket m_ReceiveReplySOCKET;
        public byte[] m_szSendBuf;
        public byte[] m_szRecvBuf;
        IPEndPoint m_SendEndPoint;      // The address and port to send to
        IPEndPoint m_localIPEndPoint;   // Local IP address and port (for receiving)

        static void Main(string[] args)
        {
	        Console.WriteLine("Local Inter-Process Communication (sender side)");
	        Console.WriteLine("This program waits for the user to type a message,");
	        Console.WriteLine("sends it to port 8007 on the local computer, and");
	        Console.WriteLine("waits for a reply message to be sent back on port 8008.");
	        Console.WriteLine("------------------------------------------------------");
	        Console.WriteLine("Type a  message to send ...   ");
 
            C_IPC_socket_Sender IPC_socket_Sender = new C_IPC_socket_Sender();
        }

        C_IPC_socket_Sender()
        {
            bool bSuccess;
            string sMessageString = Console.ReadLine(); // Get a string which may include spaces

	        // Create Send socket
	        bSuccess = CreateSendSocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }

	        // Create receive-reply socket
	        bSuccess = CreateReceiveReplySocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }
	
	        // Set the Send Address details (Use local loopback IP address (127.0.0.1), and port 8007)
	        SetUpSendAddressStructFor_LocalLoopback();
	
	        // Set up the local address structure used with bind() so the reply message can be received
	        SetUpLocalAddressStruct();

	        // Bind to local address and port 8008 so reply message can be received
	        BindToLocalAddress();

	        // Send the message that the user typed, to the sepcified port on the local computer
	        SendMessage(sMessageString);
	        Console.WriteLine("   Message sent" );
            Console.WriteLine("   Waiting for reply ...");

	        // Message has been sent - now wait for the reply
	        sMessageString = ReadMessageFromReceiveBufferOrWait();
	        Console.WriteLine("   Reply message received:    {0}", sMessageString);
	
	        CloseSocketAndExit();
        }	

        bool CreateSendSocket()
        {
	        try
            {
                m_IPC_SendSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }

        bool CreateReceiveReplySocket()
        {
	        try
            {
                m_ReceiveReplySOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }
 
        void SetUpSendAddressStructFor_LocalLoopback()
        {
	        m_SendEndPoint = new System.Net.IPEndPoint(IPAddress.Loopback, System.Convert.ToInt16(SEND_PORT));
        }
        
        void SetUpLocalAddressStruct()
        {   // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(RECEIVE_PORT));
        }

        bool BindToLocalAddress()
        {   
            try
            {   // Bind to the local IP Address and selected port
                m_ReceiveReplySOCKET.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
		        Console.WriteLine("bind() Failed!");
                return false;
            }
        }
    
        void SendMessage(string sSendString)
        {
             // Convert sSendString into the reply byte array format (!C# chars are two bytes each!)
            m_szSendBuf = Encoding.ASCII.GetBytes(sSendString);
 
            try
            {
                
                m_IPC_SendSOCKET.SendTo(m_szSendBuf, m_SendEndPoint);
            }       
            catch (SocketException ex)
            {
		        Console.WriteLine("sendto() Failed!");
		        CloseSocketAndExit();
	        }
        }
      
        string ReadMessageFromReceiveBufferOrWait()
        {
            string sReceivedStr = "";
            int iReceiveByteCount;
            try
            {
                m_szRecvBuf = new byte[RECEIVE_BUFFER_SIZE];
                IPEndPoint FromEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
                EndPoint FromAddress = (EndPoint)FromEndPoint;
                iReceiveByteCount = m_ReceiveReplySOCKET.ReceiveFrom(m_szRecvBuf, ref FromAddress);
                
                // Place the sender's address into the m_SenderEndPoint, from the SenderAddress EndPoint
                FromEndPoint.Address = IPAddress.Parse(((IPEndPoint)FromAddress).Address.ToString());
                sReceivedStr = enc8.GetString(m_szRecvBuf, 0, iReceiveByteCount); // Convert the received message into string format
            }
            catch (SocketException ex)
            {
                Console.WriteLine("Receive failed");
                CloseSocketAndExit();
            }
            return sReceivedStr;
        }

        void CloseSocketAndExit()
        {
            m_IPC_SendSOCKET.Close();
            m_ReceiveReplySOCKET.Close();
 	        Console.WriteLine("Socket closed");
	        Console.WriteLine("(press a key to continue)");
	        Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            System.Environment.Exit(1);
        }
    }
}