﻿/*	Application		Socket-based IPC threads example	
	File			Program.cs (Multithreaded_IPC)
	Purpose			Demonstration of local IPC, and multithreading using a blocking socket
	Author			Richard John Anthony
	Date			January 2015
	
	Usage			Requires two command-line arguments {send port, recv port}

	Function		1. Creates a pair of sockets (one for receiving, one for sending)
					2. Binds to the user-provided receive-port number
					3. Starts a worker thread to handle receive activity
					4. Handles send activity on the main thread
					5. Exits once the string "Quit" (or "QUIT", "quit") is entered
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;using System.Collections.Generic;
using System.Threading;

namespace Multithreaded_IPC
{
    class CMultithreaded_IPC
    {
        public const Int32 RECEIVE_BUFFER_SIZE = 100;
        public const Int32 SEND_BUFFER_SIZE = 100;
        static Encoding enc8 = Encoding.UTF8;

        Socket m_IPC_SendSOCKET;
        Socket m_IPC_RecvSOCKET;
        public byte[] m_szSendBuf;
        public byte[] m_szRecvBuf;
        int m_iSendPort;
        int m_iReceivePort;
        IPEndPoint m_SendEndPoint;      // The address and port to send to
        IPEndPoint m_localIPEndPoint;   // Local IP address and port (for receiving)
        
        static bool m_bMainThreadRunning;

        static void Main(string[] args)
        {
            if (2 != args.Length)
            {   // User did not provide the required parameters
                string sProgramName = Environment.GetCommandLineArgs()[0];  // Not the same as args[0]
		        Console.WriteLine("Usage: {0} SendPort_number ReceivePort_number", sProgramName);
		        Console.ReadKey();	// Wait for user to type a key (allow them to read the message before the window is closed)
                System.Environment.Exit(1);
            }
    
	        int iSendPort = Convert.ToInt32(args[0]);
	        if(1023 >= iSendPort)
	        {   // Send Port value is in reserved port range
		        Console.WriteLine("Send port number must be 1024 or greater (up to 65535)");
		        Console.ReadKey();	// Wait for user to type a key (allow them to read the message before the window is closed)
                System.Environment.Exit(1);
	        }
	
	        int iReceivePort = Convert.ToInt32(args[1]);
	        if(1023 >= iReceivePort)
	        {   // Receive port value is in reserved port range
		        Console.WriteLine("Receive port number must be 1024 or greater (up to 65535)");
		        Console.ReadKey();	// Wait for user to type a key (allow them to read the message before the window is closed)
                System.Environment.Exit(1);
	        }

	        Console.WriteLine("Multithreaded Local Inter-Process Communication");
            Console.WriteLine("This process sends messages on port {0} and receives messages on port {1}", iSendPort, iReceivePort);
	        Console.WriteLine("Sending is performed on main thread, receiving is performed by a worker thread");
            Console.WriteLine("------------------------------------------------------------------------------");
    
            CMultithreaded_IPC Multithreaded_IPC = new CMultithreaded_IPC(iSendPort, iReceivePort);
        }

	    CMultithreaded_IPC(int iSendPort, int iReceivePort)
        {
	        bool bSuccess;
	        m_bMainThreadRunning = false;
            m_iSendPort = iSendPort;
            m_iReceivePort = iReceivePort;

	        // Create the socket that will be used for sending messages
	        bSuccess = CreateSendSocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }

	        // Create a socket that will be used for receiving messages
	        bSuccess = CreateReceiveSocket();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting the program because a socket could not be created");
		        CloseSocketAndExit();
	        }

            SetUpLocalAddressStruct();  // Set the Local Address details (IP address and port number)

	        // Bind (The operating system makes a mapping between this process and the requested Port)
	        bSuccess = BindToLocalAddress();
	        if(false == bSuccess)
	        {
		        Console.WriteLine("Exiting because binding to the requested port failed (port may be in use)");
		        CloseSocketAndExit();
	        }

            SetUpSendAddressStruct(); // Set the Address details of the process to send to (IP address and port number)

	        m_bMainThreadRunning = true;

	        // Create and start worker thread (performs Receive in loop)
 	        Thread ReceiveThread = new Thread(ReceiveMessages);
	        ReceiveThread.Start(); // The ReceiveThread runs asynchronously with the main() thread

	        // The Send thread simply continues from here (it is the original 'main' thread of execution)
	        while(true)
	        {
 		        Console.WriteLine("SendThread: Type a message to send (\"Quit\" to quit): ");
		        string sIn = Console.ReadLine();
		        if("QUIT" == sIn.ToUpper())
		        {
			        m_bMainThreadRunning = false;		// Signal to Receive Thread to stop its operation
			        Console.WriteLine("Quiting");
			        CloseSocketAndExit();
		        }
		        int iNumberOfBytesSent = SendMessage(sIn);
		        Console.WriteLine("SendThread  Sent: {0}   number of bytes sent: {1}", sIn, iNumberOfBytesSent);
	        }
        }
	 
        bool CreateSendSocket()
        {
	        try
            {
                m_IPC_SendSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }

        bool CreateReceiveSocket()
        {
	        try
            {
                m_IPC_RecvSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
			{
 		        Console.WriteLine("CreateSocket() failed");
                return false;
            }                
            return true;
        }
 
        void SetUpSendAddressStruct()
        {
	        m_SendEndPoint = new System.Net.IPEndPoint(IPAddress.Loopback, System.Convert.ToInt16(m_iSendPort));
        }
        
        void SetUpLocalAddressStruct()
        {   // Create an Endpoint that will cause the listening activity to apply to all the local node's interfaces
            m_localIPEndPoint = new System.Net.IPEndPoint(IPAddress.Any, System.Convert.ToInt16(m_iReceivePort));
        }

        bool BindToLocalAddress()
        {   
            try
            {   // Bind to the local IP Address and selected port
                m_IPC_RecvSOCKET.Bind(m_localIPEndPoint);
                return true;
            }
            catch   // Catch any exceptions
            {
		        Console.WriteLine("bind() Failed!");
                return false;
            }
        }
  
        void CloseSocketAndExit()
        {
            m_IPC_SendSOCKET.Close();
 	        Console.WriteLine("Socket closed");
            Console.WriteLine("(press a key to continue)");
	        Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            m_IPC_RecvSOCKET.Close(); // Allow some extra time before closing (for receive thread to terminate)
            System.Environment.Exit(1);
        }

        // ******************************* Beginning of Receive Thread code *******************************
        public void ReceiveMessages(object config) // The Receive Thread body
        {
            Console.WriteLine("ReceiveThread: Waiting for message ...   ");

            while (true == CMultithreaded_IPC.m_bMainThreadRunning)
            {
                ReadMessageFromReceiveBufferOrWait();
            }
        }

        void ReadMessageFromReceiveBufferOrWait()
        {
            string sReceivedStr = "";
            int iReceiveByteCount;
            try
            {
                m_szRecvBuf = new byte[RECEIVE_BUFFER_SIZE];
                IPEndPoint FromEndPoint = new IPEndPoint(IPAddress.Any, 0);   // Initialise (will be overwritten)
                EndPoint FromAddress = (EndPoint)FromEndPoint;
                iReceiveByteCount = m_IPC_RecvSOCKET.ReceiveFrom(m_szRecvBuf, ref FromAddress);

                // Place the sender's address into the m_SenderEndPoint, from the SenderAddress EndPoint
                FromEndPoint.Address = IPAddress.Parse(((IPEndPoint)FromAddress).Address.ToString());
                sReceivedStr = enc8.GetString(m_szRecvBuf, 0, iReceiveByteCount); // Convert the received message into string format
                Console.WriteLine("RecvThread  Received: {0}   number of bytes received: {1}", sReceivedStr, iReceiveByteCount);
            }
            catch (SocketException ex)
            {
                Console.WriteLine("Receive failed");
                CloseSocketAndExit();
            }
        }
        // ********************************* End of Receive Thread code *********************************

        int SendMessage(string sSendString)
        {
            // Convert sSendString into the reply byte array format (!C# chars are two bytes each!)
            m_szSendBuf = Encoding.ASCII.GetBytes(sSendString);
            int iBytesSent = 0;
            try
            {
                iBytesSent = m_IPC_SendSOCKET.SendTo(m_szSendBuf, m_SendEndPoint);
            }       
            catch (SocketException ex)
            {
		        Console.WriteLine("sendto() Failed!");
		        CloseSocketAndExit();
	        }
	        return iBytesSent;
        }
    }
}